/*
 * Exploring Homography Estimation
 * Keith J. O'Hara <kohara@bard.edu>
 * October 2014
 * 
 * This sketch relates points in two images by a 3x3 homography matrix. 
 * You can snap pictures by pressing a key. The homography is then estimated using 
 * four points (really eight, four in each image) in the images by clicking on 
 * matching locations.   The resulting image shows both images in the reference 
 * frame of the first image using backward warping.
 * 
 * (1) First, understand how this code works.
 *
 * (2) Create a function that warps the image using forward mapping rather than
 *     backward mapping and compare the two methods. Which works better? Why?
 * 
 * (3) Create a modified sketch that allows the user to insert the webcam
 *     onto an arbitrary plane in a static image. You will need to:
 * 
 *   - use video and a static image from a file, rather than two static 
 *     images from the webcam
 *  
 *   OR
 * 
 * (3) stitch together more than two images 
 *     - either multiple pictures of a planar surface or 
 *       pictures taken from a rotated about a central point
 * 
 */

import Jama.*;
import processing.video.*;

boolean debug = false;

Capture img;
int MAXIMGS = 2;

PImage[] imgs = new PImage[MAXIMGS];

// points in our left image
PVector p1[] = new PVector[100];
int np1 = 0;

// points in our right image
PVector p2[] = new PVector[100];
int np2 = 0;

int num = 0;

Matrix homography;

void setup() {
  size(640, 480);
  img = new Capture(this, width/2, height/2);
  img.start();
  noStroke();
  background(255);

  smooth();
  homography = Matrix.identity(3, 3);
  textFont(createFont("FFScala", 32));
}

PVector applyTransformation(Matrix h, PVector v) {
  Matrix u  = new Matrix(3, 1);
  u.set(0, 0, v.x);
  u.set(1, 0, v.y);
  u.set(2, 0, 1);
  Matrix t = h.times(u);
  return new PVector((float)(t.get(0, 0)/t.get(2, 0)), (float)(t.get(1, 0)/ (t.get(2, 0))));
} 

void keyPressed() {
  imgs[num%MAXIMGS] = img.get();
  num++;
}

void mousePressed() {
  if (mouseY > imgs[0].height) {
    int i = int(mouseX / (imgs[0].width)); 
    if (i == 0) {
      p1[np1] = new PVector(mouseX % (imgs[0].width), mouseY % (imgs[0].height), 1);
      np1++;
      println("pic 1" + p1[np1-1]);
    } else if (i == 1) {
      p2[np2] = new PVector(mouseX % (imgs[0].width), mouseY % (imgs[0].height), 1);
      np2++;
      println("pic 2" + p2[np2-1]);
    }

    if (np1 >=4 && np2 >= 4 && np1 == np2) {
      estimate();
    }
  }
}

void estimate() {
  if (np1 >= 4 && np2 >= 4 && np1 == np2) {

    double[][] a = new double[2*np1][];

    // Creates the estimation matrix                                                                                                                                         
    for (int i = 0; i < np1; i++) {
      double l1 [] = {
        p2[i].x, p2[i].y, p2[i].z, 0, 0, 0, -p2[i].x*p1[i].x, -p2[i].y*p1[i].x, -p1[i].x
      };
      double l2 [] = {
        0, 0, 0, p2[i].x, p2[i].y, p2[i].z, -p2[i].x*p1[i].y, -p2[i].y*p1[i].y, -p1[i].y
      };
      a[2*i] = l1;
      a[2*i+1] = l2;
    }

    Matrix A = new Matrix(a);
    Matrix X = A.transpose().times(A);

    // assumes the eigenvalues/vectors are sorted in increasing order
    EigenvalueDecomposition E = X.eig();  
    Matrix v = E.getV();

    if (debug) {
      double[] eigenvalues = E.getRealEigenvalues();
      println(eigenvalues);
      v.print(9, 9);
    }

    // create the homography matrix from the smallest eigenvector                                                                                                                
    for (int i = 0; i < 3; i++) {
      for (int j = 0; j < 3; j++) {
        homography.set(i, j, v.get(i*3+j, 0));
      }
    }
  }

  println("Estimated H");
  homography.print(3, 3);
}

void draw() { 

  if (img.available()) {
    img.read();
  }

  if (np1 >= 4 && np2 >=4) {
    Matrix hInv = homography.inverse();

    loadPixels();
    // perform backward mapping
    for (int x = 0; x < width; x++) {
      for (int y = 0; y < height/2; y++) {

        int idx = y* width + x;

        color pix0 = color(255);
        float w0 = 0;
        if (x > 0 && x < img.width && y > 0 && y < img.height) {
          int idx0 = y* img.width + x;
          pix0 = imgs[0].pixels[idx0];
          w0 = 1;
        }

        color pix1 = color(255);
        float w1 = 0;
        PVector np = applyTransformation(hInv, new PVector(x, y, 1));
        int x1 = (int)np.x;
        int y1 = (int)np.y;
        if (x1 > 0 && x1< img.width && y1 > 0 && y1 < img.height) {
          int idx1 = y1* img.width + x1;
          pix1 = imgs[1].pixels[idx1];
          w1 = 1;
        }
        if (w0 + w1 > 0) {
          color nc = color(
          (w0*red(pix0) + w1*red(pix1)) / (w0 + w1), 
          (w0*green(pix0) + w1*green(pix1)) / (w0 + w1), 
          (w0*blue(pix0) + w1*blue(pix1)) / (w0 + w1));

          pixels[idx] = nc;
        }
      }
    }
    updatePixels();
  }

  // draw the live image
  image(img, num*(img.width), img.height, img.width, img.height);
  
  // draw the recorded images
  for (int i = 0; i < num; i++) {
    image(imgs[i], i*imgs[0].width, img.height, imgs[0].width, imgs[0].height);
  }
  // draw the selected points in image 1
  for (int i = 0; i < np1; i++) {
    fill(255, 0, 0);
    ellipse(     p1[i].x, img.height+p1[i].y, 2, 2);
    text(str(i), p1[i].x, img.height+p1[i].y);
  }
  // draw the selected points in image 2
  for (int i = 0; i < np2; i++) {
    fill(255, 0, 0);
    ellipse(     img.width + p2[i].x, img.height+p2[i].y, 2, 2);
    text(str(i), img.width + p2[i].x, img.height+p2[i].y);
  }
}

